import os
import argparse
import time
import torch
from model import SingleViewto3D
from r2n2_custom import R2N2
from  pytorch3d.datasets.r2n2.utils import collate_batched_R2N2
import dataset_location
import pytorch3d
from pytorch3d.ops import sample_points_from_meshes
from pytorch3d.ops import knn_points
import mcubes
import utils_vox


# added by yenchi
import imageio
import torch.nn.functional as F
from pytorch3d.structures import Pointclouds
from torchvision.utils import save_image

from train_model import (
        init_renderer_and_camera,
    )
from utils_yc import vox_to_mesh, add_mesh_textures, get_mesh_renderer, get_points_renderer

from tqdm import tqdm


def get_args_parser():
    parser = argparse.ArgumentParser('Singleto3D', add_help=False)
    parser.add_argument('--arch', default='resnet18', type=str)
    parser.add_argument('--max_iter', default=10000, type=str)
    parser.add_argument('--vis_freq', default=50, type=str)
    parser.add_argument('--batch_size', default=1, type=str)
    parser.add_argument('--num_workers', default=0, type=str)
    parser.add_argument('--type', default='vox', choices=['vox', 'point', 'mesh', 'vox-implicit'], type=str)
    parser.add_argument('--n_points', default=5000, type=int)
    parser.add_argument('--w_chamfer', default=1.0, type=float)
    parser.add_argument('--w_smooth', default=0.1, type=float)  
    parser.add_argument('--load_checkpoint', action='store_true')
    parser.add_argument('--ckpt_path', type=str)
    return parser

def preprocess(feed_dict):
    for k in ['images']:
        feed_dict[k] = feed_dict[k].cuda()

    images = feed_dict['images'].squeeze(1)
    mesh = feed_dict['mesh']

    return images, mesh

def compute_sampling_metrics(pred_points, gt_points, thresholds= [0.01, 0.02, 0.03, 0.04, 0.05], eps=1e-8):
    metrics = {}
    lengths_pred = torch.full(
        (pred_points.shape[0],), pred_points.shape[1], dtype=torch.int64, device=pred_points.device
    )
    lengths_gt = torch.full(
        (gt_points.shape[0],), gt_points.shape[1], dtype=torch.int64, device=gt_points.device
    )

    # For each predicted point, find its neareast-neighbor GT point
    knn_pred = knn_points(pred_points, gt_points, lengths1=lengths_pred, lengths2=lengths_gt, K=1)
    # Compute L1 and L2 distances between each pred point and its nearest GT
    pred_to_gt_dists2 = knn_pred.dists[..., 0]  # (N, S)
    pred_to_gt_dists = pred_to_gt_dists2.sqrt()  # (N, S)

    # For each GT point, find its nearest-neighbor predicted point
    knn_gt = knn_points(gt_points, pred_points, lengths1=lengths_gt, lengths2=lengths_pred, K=1)
    # Compute L1 and L2 dists between each GT point and its nearest pred point
    gt_to_pred_dists2 = knn_gt.dists[..., 0]  # (N, S)
    gt_to_pred_dists = gt_to_pred_dists2.sqrt()  # (N, S)

    # Compute precision, recall, and F1 based on L2 distances
    for t in thresholds:
        precision = 100.0 * (pred_to_gt_dists < t).float().mean(dim=1)
        recall = 100.0 * (gt_to_pred_dists < t).float().mean(dim=1)
        f1 = (2.0 * precision * recall) / (precision + recall + eps)
        metrics["Precision@%f" % t] = precision
        metrics["Recall@%f" % t] = recall
        metrics["F1@%f" % t] = f1

    # Move all metrics to CPU
    metrics = {k: v.cpu() for k, v in metrics.items()}
    return metrics

def evaluate(predictions, mesh_gt, args):
    if args.type == "vox" or args.type == 'vox-implicit':
        voxels_src = predictions
        H,W,D = voxels_src.shape[2:]
        vertices_src, faces_src = mcubes.marching_cubes(voxels_src.detach().cpu().squeeze().numpy(), isovalue=0.5)
        vertices_src = torch.tensor(vertices_src).float()
        faces_src = torch.tensor(faces_src.astype(int))
        mesh_src = pytorch3d.structures.Meshes([vertices_src], [faces_src])
        pred_points = sample_points_from_meshes(mesh_src, args.n_points)
        pred_points = utils_vox.Mem2Ref(pred_points, H, W, D)
    elif args.type == "point":
        pred_points = predictions.cpu()
    elif args.type == "mesh":
        pred_points = sample_points_from_meshes(predictions, args.n_points).cpu()

    gt_points = sample_points_from_meshes(mesh_gt, args.n_points)
        
    metrics = compute_sampling_metrics(pred_points, gt_points)
    return metrics


def render_and_save(args, step, images_gt, predictions, mesh_gt, 
                    renderer, cameras,
                    renderer_mesh_gt, cameras_mesh_gt, name=None
                    ):

    device = predictions.device

    nimgs_to_render = args.batch_size
    if args.type == 'vox' or args.type == 'vox-implicit':
        try:
            mesh_pred = vox_to_mesh(torch.sigmoid(predictions).squeeze(1))
            rend_pred = renderer(mesh_pred, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
        except:
            rend_pred = torch.zeros(nimgs_to_render, 3, 256, 256).to(predictions)
    elif args.type == 'point':
        
        pcd_src = Pointclouds(points=predictions, features=torch.ones_like(predictions))
        rend_pred = renderer(pcd_src, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
    else:
        mesh_pred = add_mesh_textures(predictions)
        rend_pred = renderer(mesh_pred, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
    
    # render of gt mesh
    mesh_gt = add_mesh_textures(mesh_gt).to(device)
    rend_gt = renderer_mesh_gt(mesh_gt, cameras=cameras_mesh_gt)[..., :3].permute(0, 3, 1, 2)

    # save them
    img_gt = F.interpolate(images_gt.permute(0, 3, 1, 2), size=(256, 256))
    comp = torch.cat([img_gt, rend_pred, rend_gt], dim=-1)
    if name == None:
        save_image(comp, f'vis/{step:03d}_{args.type}_{args.ckpt_setting}.png')
    else:
        save_image(comp, f'{name}')

def evaluate_model(args):
    r2n2_dataset = R2N2("test", dataset_location.SHAPENET_PATH, dataset_location.R2N2_PATH, dataset_location.SPLITS_PATH, return_voxels=True)

    # ============= added by yc. for visualization =============
    # init renderer
    device = 'cuda'
    if args.type == 'vox' or args.type == 'vox-implicit':
        renderer = get_mesh_renderer(image_size=256)
        dist, elev, azim = 2.7, 20, 20
    elif args.type == 'point':
        renderer = get_points_renderer(image_size=256, radius=0.01, background_color=(1,1,1))
        dist, elev, azim = 1.2, 20, 110
    else:
        renderer = get_mesh_renderer(image_size=256)
        dist, elev, azim = 1.1, 20, 110

    R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
    cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)
    # renderer, cameras = init_renderer_and_camera(args, device)

    renderer_mesh_gt = get_mesh_renderer(image_size=256)
    dist, elev, azim = 1.1, 20, 110
    R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
    cameras_mesh_gt = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)

    vis_dir = 'vis'
    if not os.path.exists(vis_dir): os.makedirs(vis_dir)

    ckpt_setting = args.ckpt_path.split('/')[1]
    args.ckpt_setting = ckpt_setting

    # ============= added by yc. for visualization =============

    loader = torch.utils.data.DataLoader(
        r2n2_dataset,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        collate_fn=collate_batched_R2N2,
        pin_memory=True,
        drop_last=True)
    eval_loader = iter(loader)

    model =  SingleViewto3D(args)
    model.cuda()
    model.eval()

    start_iter = 0
    start_time = time.time()

    avg_f1_score = []

    if args.load_checkpoint:
        # checkpoint = torch.load(f'checkpoint_{args.type}.pth')
        checkpoint = torch.load(args.ckpt_path)
        model.load_state_dict(checkpoint['model_state_dict'])
        print(f"Succesfully loaded iter {start_iter}")
    
    print("Starting evaluating !")
    max_iter = len(eval_loader)
    pbar = tqdm(total=max_iter)
    for step in range(start_iter, max_iter):
        # if step == 5:
            # break

        iter_start_time = time.time()

        read_start_time = time.time()

        feed_dict = next(eval_loader)

        images_gt, mesh_gt = preprocess(feed_dict)

        read_time = time.time() - read_start_time

        predictions = model(images_gt, args)

        if args.type == "vox" or args.type == 'vox-implicit':
            predictions = predictions.permute(0,1,4,3,2)
            try:
                metrics = evaluate(torch.sigmoid(predictions), mesh_gt, args)
            except:
                metrics = {}
                metrics['F1@0.050000'] = torch.tensor([0.])
        else:
            metrics = evaluate(predictions, mesh_gt, args)

        # TODO:
        if (step % args.vis_freq) == 0:
            # visualization block
            #  rend = 
            # plt.imsave(f'vis/{step}_{args.type}.png', rend)
            render_and_save(args, step, images_gt, predictions, mesh_gt, 
                    renderer, cameras,
                    renderer_mesh_gt, cameras_mesh_gt)
            

        total_time = time.time() - start_time
        iter_time = time.time() - iter_start_time

        f1_05 = metrics['F1@0.050000']

        avg_f1_score.append(f1_05)

        # print("[%4d/%4d]; ttime: %.0f (%.2f, %.2f); F1@0.05: %.3f; Avg F1@0.05: %.3f" % (step, max_iter, total_time, read_time, iter_time, f1_05, torch.tensor(avg_f1_score).mean()))
        msg = "[%4d/%4d]; ttime: %.0f (%.2f, %.2f); F1@0.05: %.3f; Avg F1@0.05: %.3f" % (step, max_iter, total_time, read_time, iter_time, f1_05, torch.tensor(avg_f1_score).mean())
        pbar.set_description(msg)
        pbar.update(1)

    with open('eval-f1.txt', 'a') as f:
        f.write(f'{args.type}-{ckpt_setting}: Avg F1@0.05: {torch.tensor(avg_f1_score).mean()}\n')

    print('Done!')

if __name__ == '__main__':
    parser = argparse.ArgumentParser('Singleto3D', parents=[get_args_parser()])
    args = parser.parse_args()
    evaluate_model(args)
