import argparse
import os
import pdb
import time

import losses
from pytorch3d.utils import ico_sphere
from r2n2_custom import R2N2
from pytorch3d.ops import sample_points_from_meshes
from pytorch3d.structures import Meshes, Pointclouds
import dataset_location
import torch

# added by yenchi
import imageio
import numpy as np
import pytorch3d

from tqdm import tqdm
from utils_yc import (
    get_mesh_renderer,
    get_points_renderer,
    add_mesh_textures,
    vox_to_mesh,
    plot_scene_to_html
)

def get_args_parser():
    parser = argparse.ArgumentParser('Model Fit', add_help=False)
    parser.add_argument('--lr', default=4e-4, type=float)
    parser.add_argument('--max_iter', default=100000, type=int)
    parser.add_argument('--log_freq', default=1000, type=int)
    parser.add_argument('--type', default='vox', choices=['vox', 'point', 'mesh'], type=str)
    parser.add_argument('--n_points', default=5000, type=int)
    parser.add_argument('--w_chamfer', default=1.0, type=float)
    parser.add_argument('--w_smooth', default=0.1, type=float)
    return parser

def fit_mesh(mesh_src, mesh_tgt, args):
    device = mesh_src.device
    start_iter = 0
    start_time = time.time()

    deform_vertices_src = torch.zeros(mesh_src.verts_packed().shape,requires_grad=True,device='cuda')
    optimizer = torch.optim.Adam([deform_vertices_src], lr = args.lr)
    print("Starting training !")

    pbar = tqdm(total=args.max_iter)
    for step in range(start_iter, args.max_iter):
        iter_start_time = time.time()

        new_mesh_src = mesh_src.offset_verts(deform_vertices_src)

        sample_trg = sample_points_from_meshes(mesh_tgt, args.n_points)
        sample_src = sample_points_from_meshes(new_mesh_src, args.n_points)

        loss_reg = losses.chamfer_loss(sample_src, sample_trg)
        loss_smooth = losses.smoothness_loss(new_mesh_src)

        loss = args.w_chamfer * loss_reg + args.w_smooth * loss_smooth

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()        

        total_time = time.time() - start_time
        iter_time = time.time() - iter_start_time

        loss_vis = loss.cpu().item()

        # print("[%4d/%4d]; ttime: %.0f (%.2f); loss: %.3f" % (step, args.max_iter, total_time,  iter_time, loss_vis))    
        msg = "[%4d/%4d]; ttime: %.0f (%.2f); loss: %.3f" % (step, args.max_iter, total_time,  iter_time, loss_vis)
        pbar.set_description(msg)
        pbar.update(1)    
    
    mesh_src.offset_verts_(deform_vertices_src)

    # add textures
    mesh_src = add_mesh_textures(mesh_src)
    mesh_tgt = add_mesh_textures(mesh_tgt)

    renderer = get_mesh_renderer(image_size=256)

    dist = 1.1
    elev = 20
    azim = 110
    R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
    cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)

    img_src = renderer(mesh_src, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()
    img_tgt = renderer(mesh_tgt, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()

    img_src = (img_src * 255.).astype(np.uint8)
    img_tgt = (img_tgt * 255.).astype(np.uint8)

    # plot_mesh_cams_scene(mesh_tgt, cameras)

    # QQQQ: weird textures
    out_dir = 'output'
    imageio.imwrite(f"{out_dir}/q1.3-mesh-src.png", img_src)
    imageio.imwrite(f"{out_dir}/q1.3-mesh-tgt.png", img_tgt)

    print('Done!')

def fit_pointcloud(pointclouds_src, pointclouds_tgt, args):
    device = pointclouds_src.device
    start_iter = 0
    start_time = time.time()    
    optimizer = torch.optim.Adam([pointclouds_src], lr = args.lr)

    pbar = tqdm(total=args.max_iter)

    for step in range(start_iter, args.max_iter):
        # break
        iter_start_time = time.time()

        loss = losses.chamfer_loss(pointclouds_src, pointclouds_tgt)
        # loss, _ = pytorch3d.loss.chamfer_distance(pointclouds_src, pointclouds_tgt)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()        

        total_time = time.time() - start_time
        iter_time = time.time() - iter_start_time

        loss_vis = loss.cpu().item()

        msg = "[%4d/%4d]; ttime: %.0f (%.2f); loss: %.4f" % (step, args.max_iter, total_time,  iter_time, loss_vis)
        pbar.set_description(msg)
        pbar.update(1)
    
    # vis - render voxel
    renderer = get_points_renderer(image_size=256)

    dist = 1.2
    elev = 20
    azim = 110 ######## Q. what is the reference vector (0, 0, 1)? it is pointing out from the screen.
    R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
    cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)

    pcd_src = Pointclouds(points=pointclouds_src, features=torch.ones_like(pointclouds_src))
    pcd_tgt = Pointclouds(points=pointclouds_tgt, features=torch.ones_like(pointclouds_tgt))

    img_src = renderer(pcd_src, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()
    img_tgt = renderer(pcd_tgt, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()

    img_src = (img_src * 255.).astype(np.uint8)
    img_tgt = (img_tgt * 255.).astype(np.uint8)

    # plot_mesh_cams_scene(mesh_tgt, cameras) # for debug

    out_dir = 'output'
    imageio.imwrite(f"{out_dir}/q1.2-pcd-src.png", img_src)
    imageio.imwrite(f"{out_dir}/q1.2-pcd-tgt.png", img_tgt)

    print('Done!')


def fit_voxel(voxels_src, voxels_tgt, args):
    device = voxels_src.device

    start_iter = 0
    start_time = time.time()    
    optimizer = torch.optim.Adam([voxels_src], lr = args.lr)

    pbar = tqdm(total=args.max_iter)

    for step in range(start_iter, args.max_iter):
        iter_start_time = time.time()

        loss = losses.voxel_loss(voxels_src, voxels_tgt)

        optimizer.zero_grad()
        loss.backward(retain_graph=True)
        optimizer.step()

        total_time = time.time() - start_time
        iter_time = time.time() - iter_start_time

        loss_vis = loss.cpu().item()

        # print("[%4d/%4d]; ttime: %.0f (%.2f); loss: %.6f" % (step, args.max_iter, total_time,  iter_time, loss_vis))
        msg = "[%4d/%4d]; ttime: %.0f (%.2f); loss: %.6f" % (step, args.max_iter, total_time,  iter_time, loss_vis)
        pbar.set_description(msg)
        pbar.update(1)
        # break

    # vis - render voxel
    renderer = get_mesh_renderer(image_size=256)

    dist = 2.7
    elev = 20
    azim = 110 ######## Q. what is the reference vector (0, 0, 1)? it is pointing out from the screen.
    R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
    cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)

    mesh_src = vox_to_mesh(voxels_src)
    mesh_tgt = vox_to_mesh(voxels_tgt)

    img_src = renderer(mesh_src, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()
    img_tgt = renderer(mesh_tgt, cameras=cameras)[0, :, :, :3].detach().cpu().numpy()

    img_src = (img_src * 255.).astype(np.uint8)
    img_tgt = (img_tgt * 255.).astype(np.uint8)

    # plot_mesh_cams_scene(mesh_tgt, cameras)

    out_dir = 'output'
    imageio.imwrite(f"{out_dir}/q1.1-vox-src.png", img_src)
    imageio.imwrite(f"{out_dir}/q1.1-vox-tgt.png", img_tgt)

    print('Done!')

def train_model(args):
    r2n2_dataset = R2N2("train", dataset_location.SHAPENET_PATH, dataset_location.R2N2_PATH, dataset_location.SPLITS_PATH, return_voxels=True)
    
    feed = r2n2_dataset[0]

    feed_cuda = {}
    for k in feed:
        if torch.is_tensor(feed[k]):
            feed_cuda[k] = feed[k].cuda().float()


    if args.type == "vox":
        # initialization
        voxels_src = torch.rand(feed_cuda['voxels'].shape,requires_grad=True,device='cuda')
        voxel_coords = feed_cuda['voxel_coords'].unsqueeze(0)
        voxels_tgt = feed_cuda['voxels'].to(voxels_src)

        # fitting
        fit_voxel(voxels_src, voxels_tgt, args)


    elif args.type == "point":
        # initialization
        pointclouds_src = torch.randn([1,args.n_points,3],requires_grad=True,device='cuda')
        mesh_tgt = Meshes(verts=[feed_cuda['verts']], faces=[feed_cuda['faces']])
        pointclouds_tgt = sample_points_from_meshes(mesh_tgt, args.n_points)

        # fitting
        fit_pointcloud(pointclouds_src, pointclouds_tgt, args)        
    
    elif args.type == "mesh":
        # initialization
        # try different ways of initializing the source mesh        
        mesh_src = ico_sphere(4,'cuda')
        mesh_tgt = Meshes(verts=[feed_cuda['verts']], faces=[feed_cuda['faces']])

        # fitting
        fit_mesh(mesh_src, mesh_tgt, args)        


    
    


if __name__ == '__main__':
    parser = argparse.ArgumentParser('Model Fit', parents=[get_args_parser()])
    args = parser.parse_args()

    out_dir = 'output'
    if not os.path.exists(out_dir): os.makedirs(out_dir)

    train_model(args)
