import argparse
import pdb
import time
import torch
from model import SingleViewto3D
from r2n2_custom import R2N2
from  pytorch3d.datasets.r2n2.utils import collate_batched_R2N2
import dataset_location
from pytorch3d.ops import sample_points_from_meshes
import losses

# added by yc
import pytorch3d
from pytorch3d.structures import Pointclouds, Meshes
from torch.backends import cudnn
cudnn.benchmark = True

import os
from tqdm import tqdm
from einops import rearrange
from torch.utils.tensorboard import SummaryWriter
from torchvision.utils import make_grid

from utils_yc import (
    get_mesh_renderer,
    get_points_renderer,
    add_mesh_textures,
    vox_to_mesh,
    plot_scene_to_html
)

def get_args_parser():
    parser = argparse.ArgumentParser('Singleto3D', add_help=False)
    # Model parameters
    parser.add_argument('--arch', default='resnet18', type=str)
    parser.add_argument('--lr', default=4e-4, type=float)
    parser.add_argument('--max_iter', default=10000, type=str)
    parser.add_argument('--log_freq', default=1000, type=str)
    parser.add_argument('--batch_size', default=2, type=int)
    parser.add_argument('--num_workers', default=0, type=int)
    parser.add_argument('--type', default='vox', choices=['vox', 'point', 'mesh', 'vox-implicit'], type=str)
    parser.add_argument('--n_points', default=5000, type=int)
    parser.add_argument('--w_chamfer', default=1.0, type=float)
    parser.add_argument('--w_smooth', default=0.1, type=float)
    parser.add_argument('--save_freq', default=2000, type=int)
    parser.add_argument('--vis_freq', default=500, type=int)    
    parser.add_argument('--load_checkpoint', action='store_true')            
    return parser

def preprocess(feed_dict,args):
    for k in ['images','voxels','mesh']:
        feed_dict[k] = feed_dict[k].cuda()
        
    images = feed_dict['images'].squeeze(1)
    if args.type == "vox" or args.type == 'vox-implicit':
        voxels = feed_dict['voxels'].float()
        ground_truth_3d = voxels
    elif args.type == "point":
        mesh = feed_dict['mesh']
        pointclouds_tgt = sample_points_from_meshes(mesh, args.n_points)    
        ground_truth_3d = pointclouds_tgt        
    elif args.type == "mesh":
        ground_truth_3d = feed_dict["mesh"]
    return images, ground_truth_3d

def calculate_loss(predictions, ground_truth, args):
    if args.type == 'vox' or args.type == 'vox-implicit':
        loss = losses.voxel_loss(predictions,ground_truth)
    elif args.type == 'point':
        loss = losses.chamfer_loss(predictions, ground_truth)
        # loss, _ = pytorch3d.loss.chamfer_distance(predictions, ground_truth)
    elif args.type == 'mesh':
        sample_trg = sample_points_from_meshes(ground_truth, args.n_points)
        sample_pred = sample_points_from_meshes(predictions, args.n_points)

        loss_reg = losses.chamfer_loss(sample_pred, sample_trg)
        loss_smooth = losses.smoothness_loss(predictions)

        loss = args.w_chamfer * loss_reg + args.w_smooth * loss_smooth        
    return loss


def init_renderer_and_camera(args, device):
    if args.type == 'vox' or args.type == 'vox-implicit':
        dist, elev, azim = 2.7, 20, 110
        renderer = get_mesh_renderer(image_size=256)
        R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
        cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)
    elif args.type == 'point':
        renderer = get_points_renderer(image_size=256)
        dist, elev, azim = 1.2, 20, 110
        R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
        cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)
    else:
        renderer = get_mesh_renderer(image_size=256)
        dist, elev, azim = 1.1, 20, 110
        R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
        cameras = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, fov=60, device=device)
    return renderer, cameras

def render_and_vis(args, iters, writer,
                   renderer, cameras,
                   images_gt, prediction_3d, ground_truth_3d):

    nimgs_to_render = 16
    images_gt = images_gt[:nimgs_to_render]
    prediction_3d = prediction_3d[:nimgs_to_render]
    ground_truth_3d = ground_truth_3d[:nimgs_to_render]

    if args.type == 'vox' or args.type == 'vox-implicit':
        try:
            mesh_pred = vox_to_mesh(torch.sigmoid(prediction_3d).squeeze(1))
            rend_pred = renderer(mesh_pred, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
        except:
            rend_pred = torch.zeros(nimgs_to_render, 3, 256, 256).to(prediction_3d)

        mesh_gt = vox_to_mesh(ground_truth_3d.squeeze(1))
        rend_gt = renderer(mesh_gt, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
        
    elif args.type == 'point':
        
        pcd_src = Pointclouds(points=prediction_3d, features=torch.ones_like(prediction_3d))
        pcd_tgt = Pointclouds(points=ground_truth_3d, features=torch.ones_like(ground_truth_3d))
        
        rend_pred = renderer(pcd_src, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
        rend_gt = renderer(pcd_tgt, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
    
    else:
        mesh_pred = add_mesh_textures(prediction_3d)
        mesh_gt = add_mesh_textures(ground_truth_3d)
        
        # import pdb; pdb.set_trace()
        
        rend_pred = renderer(mesh_pred, cameras=cameras)[..., :3].permute(0, 3, 1, 2)
        rend_gt = renderer(mesh_gt, cameras=cameras)[..., :3].permute(0, 3, 1, 2)

    images_gt = rearrange(images_gt, 'b h w c -> b c h w')
    images_gt = make_grid(images_gt, nrow=4)
    rend_pred = make_grid(rend_pred, nrow=4)
    rend_gt = make_grid(rend_gt, nrow=4)

    writer.add_images(f'0-img', images_gt, iters, dataformats='CHW')
    writer.add_images(f'1-pred', rend_pred, iters, dataformats='CHW')
    writer.add_images(f'2-gt', rend_gt, iters, dataformats='CHW')

def train_model(args):
    device = 'cuda'
    args.device = device
    r2n2_dataset = R2N2("train", dataset_location.SHAPENET_PATH, dataset_location.R2N2_PATH, dataset_location.SPLITS_PATH, return_voxels=True)

    loader = torch.utils.data.DataLoader(
        r2n2_dataset,
        batch_size=args.batch_size,
        num_workers=args.num_workers,
        collate_fn=collate_batched_R2N2,
        pin_memory=True,
        drop_last=True)
    train_loader = iter(loader)

    model =  SingleViewto3D(args)
    model.cuda()
    model.train()

    # ============ init renderer and tensorboard writer ==============
    if args.type == 'vox':
        q = '2.1-vox'
    elif args.type == 'point':
        q = '2.2-pcd'
    elif args.type == 'vox-implicit':
        q = '3.1-vox-implicit'
    else:
        q = '2.3-mesh'

    renderer, cameras = init_renderer_and_camera(args, device)

    q = f'{q}-n_points_{args.n_points}-w_chamfer-{args.w_chamfer}'
    args.q = q

    tb_dir = f'logs/{q}/tb'
    ckpt_dir = f'logs/{q}'
    if not os.path.exists(tb_dir): os.makedirs(tb_dir)
    if not os.path.exists(ckpt_dir): os.makedirs(ckpt_dir)

    writer = SummaryWriter(log_dir=tb_dir)

    # ============ preparing optimizer ... ============
    optimizer = torch.optim.Adam(model.parameters(), lr = args.lr)  # to use with ViTs
    start_iter = 0
    start_time = time.time()

    if args.load_checkpoint:
        checkpoint = torch.load(f'checkpoint_{args.type}.pth')
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        start_iter = checkpoint['step']
        print(f"Succesfully loaded iter {start_iter}")
    
    print("Starting training !")
    pbar = tqdm(total=args.max_iter)
    for step in range(start_iter, args.max_iter):
        iter_start_time = time.time()

        if step % len(train_loader) == 0: #restart after one epoch
            train_loader = iter(loader)

        read_start_time = time.time()

        feed_dict = next(train_loader)

        images_gt, ground_truth_3d = preprocess(feed_dict,args)
        read_time = time.time() - read_start_time

        prediction_3d = model(images_gt, args)

        # import pdb; pdb.set_trace()
        loss = calculate_loss(prediction_3d, ground_truth_3d, args)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()        

        total_time = time.time() - start_time
        iter_time = time.time() - iter_start_time

        loss_vis = loss.cpu().item()

        if ((step+1) % args.save_freq) == 0:
            # torch.save({
            #     'step': step,
            #     'model_state_dict': model.state_dict(),
            #     'optimizer_state_dict': optimizer.state_dict()
            #     }, f'checkpoint_{args.type}.pth')
            torch.save({
                'step': step,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict()
                }, f'{ckpt_dir}/checkpoint_{args.type}-{step+1}.pth')

        writer.add_scalar('loss', loss_vis, step)

        if (step % args.vis_freq) == 0:
            # tboard. vis pred, output
            render_and_vis(args, step, writer, renderer, cameras, 
                           images_gt, prediction_3d, ground_truth_3d)

        # print("[%4d/%4d]; ttime: %.0f (%.2f, %.2f); loss: %.3f" % (step, args.max_iter, total_time, read_time, iter_time, loss_vis))
        msg = "[%4d/%4d]; ttime: %.0f (%.2f, %.2f); loss: %.3f" % (step, args.max_iter, total_time, read_time, iter_time, loss_vis)
        pbar.set_description(msg)
        pbar.update(1)

    print('Done!')

if __name__ == '__main__':
    parser = argparse.ArgumentParser('Singleto3D', parents=[get_args_parser()])
    args = parser.parse_args()
    train_model(args)
