import numpy as np
from scipy.misc import face
import torch
import pytorch3d

from pytorch3d.structures import Meshes
from pytorch3d.renderer import RasterizationSettings, MeshRenderer, MeshRasterizer, HardPhongShader

from pytorch3d.renderer import PointsRasterizationSettings, PointsRenderer, PointsRasterizer, AlphaCompositor

import math

# from assignment 1
def get_device():
    """
    Checks if GPU is available and returns device accordingly.
    """
    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")
    return device

def get_mesh_renderer(image_size=512, lights=None, device=None):
    """
    Returns a Pytorch3D Mesh Renderer.

    Args:
        image_size (int): The rendered image size.
        lights: A default Pytorch3D lights object.
        device (torch.device): The torch device to use (CPU or GPU). If not specified,
            will automatically use GPU if available, otherwise CPU.
    """
    if device is None:
        if torch.cuda.is_available():
            device = torch.device("cuda:0")
        else:
            device = torch.device("cpu")
    raster_settings = RasterizationSettings(
        image_size=image_size, blur_radius=0.0, faces_per_pixel=1,
    )
    renderer = MeshRenderer(
        rasterizer=MeshRasterizer(raster_settings=raster_settings),
        shader=HardPhongShader(device=device, lights=lights),
    )
    return renderer

def get_points_renderer(
    image_size=512, device=None, radius=0.01, background_color=(1, 1, 1)
):
    """
    Returns a Pytorch3D renderer for point clouds.

    Args:
        image_size (int): The rendered image size.
        device (torch.device): The torch device to use (CPU or GPU). If not specified,
            will automatically use GPU if available, otherwise CPU.
        radius (float): The radius of the rendered point in NDC.
        background_color (tuple): The background color of the rendered image.
    
    Returns:
        PointsRenderer.
    """
    if device is None:
        if torch.cuda.is_available():
            device = torch.device("cuda:0")
        else:
            device = torch.device("cpu")
    raster_settings = PointsRasterizationSettings(image_size=image_size, radius=radius,)
    renderer = PointsRenderer(
        rasterizer=PointsRasterizer(raster_settings=raster_settings),
        compositor=AlphaCompositor(background_color=background_color),
    )
    return renderer


def init_mesh(verts, faces, color=[0.7, 0.7, 1], device=None):
    if device == None:
        device = get_device()

    if verts.dim() == 2:
        verts = verts.unsqueeze(0)  # (N_v, 3) -> (1, N_v, 3)
    if faces.dim() == 2:
        faces = faces.unsqueeze(0)  # (N_f, 3) -> (1, N_f, 3)
    textures = torch.ones_like(verts)  # (1, N_v, 3)

    # color=[0.7, 0.7, 1]
    textures = textures * torch.tensor(color)  # (1, N_v, 3)

    mesh = pytorch3d.structures.Meshes(
        verts=verts,
        faces=faces,
        textures=pytorch3d.renderer.TexturesVertex(textures),
    )
    mesh = mesh.to(device)
    return mesh


def render_360(mesh, renderer, elev=0):
    device = mesh.device

    # get cameras
    dist = 3
    # elev = 0
    azim_list = torch.linspace(-180, 180, 36)

    # Place a point light in front of the cow.
    lights = pytorch3d.renderer.PointLights(location=torch.tensor([[0, 0, -3]]), device=device)

    imgs = []
    for azim in azim_list:
        R, T = pytorch3d.renderer.cameras.look_at_view_transform(dist, elev, azim)
        
        cameras = pytorch3d.renderer.FoVPerspectiveCameras(
        R=R, T=T, fov=60, device=device
        )

        # Place a point light at camera's position
        loc_light = torch.inverse(R)[0] @ -T[0]
        lights = pytorch3d.renderer.PointLights(location=loc_light.unsqueeze(0), device=device)

        rend = renderer(mesh, cameras=cameras, lights=lights)
        rend = (rend.cpu().numpy()[0, ..., :3] * 255).astype(np.uint8)

        imgs.append(rend)

        # imageio.imsave('img.png', imgs[0][0, ..., :-3].cpu().numpy())
        # import pdb; pdb.set_trace()
    return imgs


def get_cam(R0, T0, azim, device):
    R_rel = pytorch3d.transforms.axis_angle_to_matrix(torch.tensor([[0., azim * math.pi / 180., 0]])).squeeze(0)
    R = (R_rel @ R0).unsqueeze(0)
    # T = (R_rel @ T0).unsqueeze(0)
    # T = (torch.inverse(R_rel) @ T0) + T0 #
    T = T0.unsqueeze(0)
    cam = pytorch3d.renderer.FoVPerspectiveCameras(R=R, T=T, device=device)
    return cam

# =============== for assignment 2 ===============

# convert data
def vox_to_mesh(vox):

    mesh = pytorch3d.ops.cubify(vox, thresh=0.5)
    mesh = add_mesh_textures(mesh)
    return mesh

# rendering
def add_mesh_textures(mesh):
    verts = mesh.verts_list()
    faces = mesh.faces_list()
    
    bs = len(verts)
    verts_rgb = []
    for i in range(bs):
        verts_rgb.append(torch.ones_like(verts[i]))
    
    # mesh = Meshes(verts=verts, faces=faces, textures=pytorch3d.renderer.mesh.TexturesVertex(verts_features=verts_rgb))
    mesh.textures = pytorch3d.renderer.mesh.TexturesVertex(verts_rgb)
    return mesh

def render_vox(vox, renderer=None, camera=None):
    mesh = vox_to_mesh(vox)



# =============== for assignment 2 ===============

# for debugging
def plot_scene_to_html(mesh, cameras, out_f='scene.html'):
    from pytorch3d.vis.plotly_vis import plot_scene, AxisArgs

    fig = plot_scene({
        "fig": {
        "mesh": mesh,
        # "dolphin_mesh": dolphin_mesh,
        # "table_mesh": table_mesh,
        "camera": cameras,
        }
        }, axis_args=AxisArgs(showgrid=True, backgroundcolor="#ccc"))
    fig.show()
    # fig.write_html("output/q1.1-scene.html")
    fig.write_html(out_f)